﻿-- function schedluer frame
local BMTSchedulerFrame = CreateFrame("Frame");
-- time since last call of the OnUpdate
local BMTElapsed = 0;
BMTScheduledTasks = {};
-- OnUpdate calls throttle
local BMT_ONUPDATE_SEC = 0.5;
-- set to true for additional dev messages
local BMTDebug = false;
--[===[@debug@
BMTDebug = true;
--@end-debug@]===]
-- addon name
local BMT_ADDON_NAME = "BlackMarketTracker";
-- title text on top of the browser window
local BMT_TITLE = "Black Market Tracker";
-- new addon version text shown at the bottom of the browser
local BMT_NEW_VERSION = "New addon version available";
-- coming soon text
local BMT_COMING_SOON = "Coming Soon!";
-- more than 1 week ago text (last update)
local BMT_MORE_THAN_WEEK = "More than 1 week";
-- less than 1 minute ago text (last update)
local BMT_LESS_THAN_MINUTE = "Less than 1 minute";
-- completed auction text
local BMT_COMPLETED = "Completed!";
-- alpa for fading bottom text, goes above 1.0 so that the text stays lit for a while
local BMT_TEXT_ALPHA = 6.0;
-- time of fade in/out
local BMT_TEXT_TIME = 2.5;
-- sync string separator
local BMT_SEPARATOR = "%";
-- sync sender (through new search) key
local BMT_SYNC_SENDER_KEY = "UPDATER";
-- request update times key
local BMT_SYNC_TIMEREQ_KEY = "TIMEREQ";
-- response update times key
local BMT_SYNC_TIMERESP_KEY = "TIMERESP";
-- sync item key
local BMT_SYNC_ITEM_KEY = "IT";
-- sync item request key
local BMT_SYNC_ITEMREQ_KEY = "ITEMREQ";
-- sync item end key
local BMT_SYNC_ITEMEND_KEY = "ITEMEND";
-- sync version reqiest key
local BMT_SYNC_VERREQ_KEY = "VERREQ";
-- sync version response key
local BMT_SYNC_VERRESP_KEY = "VERRESP";
-- time for syncing the update times upon logon, in seconds
local BMT_SYNC_TIME_DELAY = 7;
-- time for syncing the items, in seconds
local BMT_SYNC_ITEMS_DELAY = 3;
-- time between window refreshes (TODO: This will be in the config)
local BMT_WINDOW_REFRESH_DELAY = 5;
-- maximum length of addon message. This does not include prefix or the char that delimits prefix and message - WoWWiki is wrong.
local BMT_MAX_ADDON_MSG_LEN = 255;
-- time after which auction will be removed from current db if no updates were done during that time and we got another item with same marketID
-- 48h
BMT_ITEM_REMOVE_TIME_THRESHOLD = 2880;
-- time that is added to current time left upon placing a bid on an item
-- 5m (TODO: NEEDS TESTING!)
BMT_TIME_EXTEND_PER_BID = 300;
-- predicted time left constants (time in minutes) - completed, short, medium, long, very long
BMT_TIME_LEFT = {0, 30, 120, 720, 9999};
-- array storing current db fields that are synced
BMTItemSyncFields = 	{ "name",	"texture",	"quantity",	"itemType",	"level",	"levelType",	"sellerName",	"minBid",	"minIncrement",	"currBid",	"numBids",	"timeLeft",	"link",	"predictedLow",	"predictedHigh"	};
BMTIsSyncFieldNumber =	{ false,	false,		true,		false,		true,		false,			false, 			true,		true,			true,		true,		true,		false,	true,			true			};
-- this revision number
--BMTRevision = tonumber(GetAddOnMetadata("BlackMarketTracker", "Version"));
BMTRevision = 28;
--BMTRevision = 1;
-- highest revision gotten from others upon logging in
BMTHighestRevision = BMTRevision;
-- fading state (bottom text)
BMTFadeState = 0;
-- last updated time string, used to throttle the sync results (when multiple events get fired after opening BMAH window once)
BMTLastUpdateTime = "-";
-- true if currently syncing items on logon, false otherwise. When false, BMT_SYNC_TIMERESP_KEY is ignored
BMTSyncingTime = false;
-- best sync date and time
BMTBestUpdateDate = "01/01/01";
BMTBestUpdateTime = 0;
-- name of the person with best sync time
BMTBestUpdateTimeSender = nil;
-- item database from syncing
BMTSyncItemDB = nil;
-- sender information from syncing
BMTSyncSender = nil;

function BMT_Debug(msg)
	if(BMTDebug) then
		DEFAULT_CHAT_FRAME:AddMessage(msg);
	end
end

function BMT_SlashDebugHandler()
	for i=1, #BlackMarketTrackerCurrentDB do
		local name = BlackMarketTrackerCurrentDB[i].name;
		local marketID = BlackMarketTrackerCurrentDB[i].marketID;
		BMT_Debug("name="..name.." marketID="..marketID);
	end
end

BMTSchedulerFrame:SetScript("OnUpdate", function(self, elapsedVal)
	-- throttle BMT_OnUpdate calls
	BMTElapsed = BMTElapsed + elapsedVal;
	if(BMTElapsed >= BMT_ONUPDATE_SEC) then
		BMTElapsed = 0;
		BMT_OnUpdate(self, elapsedVal);
	end
end)

function BMT_SlashDeleteHandler()
	DEFAULT_CHAT_FRAME:AddMessage("Black Market Tracker: Deleting all data.");
	BlackMarketTrackerCurrentDB = nil;
	BlackMarketTrackerHistoryDB = nil;
	BlackMarketTrackerConfig = nil;
	BlackMarketTrackerMisc = nil;
	ReloadUI();
end

function BMT_SlashHandler(msg, editbox)
	local msgLen = #msg;
	-- TODO: Config
	if(msgLen == 0 or msg == "show" or msg == "ui" or msg == "open") then
		BMTFrame_Toggle();
	elseif(msg == "debug") then
		BMT_SlashDebugHandler();
	elseif(msg == "sync") then
		BMT_OnLoginHandler();
	elseif(msg == "delete") then
		BMT_SlashDeleteHandler();
	end
end

function BMTFrame_OnLoad(self)
	SLASH_BLACKMARKETTRACKER1 = "/bmt";
	SLASH_BLACKMARKETTRACKER2 = "/blackmarkettracker";
	SlashCmdList["BLACKMARKETTRACKER"] = BMT_SlashHandler;
	self:RegisterForDrag("LeftButton");
	tinsert(UISpecialFrames, self:GetName());
	self:RegisterEvent("PLAYER_ENTERING_WORLD");
	self:RegisterEvent("CHAT_MSG_ADDON");
	self:RegisterEvent("BLACK_MARKET_ITEM_UPDATE");
	BlackMarketTrackerScrollFrame.scrollBar.doNotHide = true;
	BlackMarketTrackerFrameTitleText:SetText(BMT_TITLE);
	RegisterAddonMessagePrefix("BMT");
end

function BMTFrame_OnEvent(self, event, ...)
	if(event == "BLACK_MARKET_ITEM_UPDATE") then
		BMTTaskManager_Unschedule(BMT_NewSearchHandler);
		BMTTaskManager_Schedule(0.5, BMT_NewSearchHandler);
	elseif(event == "CHAT_MSG_ADDON") then
		local prefix = select(1, ...);
		local msg = select(2, ...);
		local channel = select(3, ...);
		local sender = select(4, ...);
		BMT_AddonMsgReceiveHandler(prefix, msg, channel, sender);
	elseif(event == "PLAYER_ENTERING_WORLD") then
		BMTTaskManager_Unschedule(BMT_OnLoginHandler);
		BMTTaskManager_Schedule(2, BMT_OnLoginHandler);
	end
end

function BMTFrame_Toggle()
	if(BlackMarketTrackerFrame:IsShown()) then
		BlackMarketTrackerFrame:Hide();
		BMTTaskManager_Unschedule(BMT_FullWindowRefresh);
		--BlackMarketTrackerBottom:SetAlpha(0);
		-- TODO: sound enable toggle
	else
		BlackMarketTrackerFrame:Show();
		-- TODO: sound enable toggle
	end
end

function BMTFrame_OnShow()
	if(BMTHighestRevision <= BMTRevision) then
		BMTFadeState = 2;
	else
		BMTFadeState = 0;
	end
	PlaySound("AuctionWindowOpen");
	HybridScrollFrame_CreateButtons(BlackMarketTrackerScrollFrame, "BlackMarketTrackerItemTemplate", 5, -5);
	BlackMarketTrackerScrollFrame_Update();
	BMTBottomText_FadingHandler();
	BMTTaskManager_Unschedule(BMT_FullWindowRefresh);
	BMTTaskManager_Schedule(BMT_WINDOW_REFRESH_DELAY, BMT_FullWindowRefresh);
end

function BMTFrame_OnHide()
	PlaySound("AuctionWindowClose");
	BMTTaskManager_Unschedule(BMT_FullWindowRefresh);
end

-- delay_time, function, function_parameters
function BMTTaskManager_Schedule(delay, func, ...)
	-- add params
	local task = {...};
	-- add function and time when it should be called
	task.func = func;
	task.time = GetTime() + delay;
	table.insert(BMTScheduledTasks, task);
end

function BMTTaskManager_Unschedule(func, ...)
	local entry, found;
	for i=#BMTScheduledTasks, 1, -1 do
		entry = BMTScheduledTasks[i];
		if(entry.func == func) then
			found = true;
			for j=1, select("#", ...) do
				if(select(j, ...) ~= entry[i]) then
					found = false;
					break;
				end
			end
			if(found) then
				table.remove(BMTScheduledTasks, i);
			end
		end
	end
end

function BMT_OnUpdate(self, elapsedVal)
	-- for some reason this needs to be done
	if(BMTElapsed == 0) then return end
	
	--BMT_Debug(GetTime().." BMTElapsed="..BMTElapsed);
	local entry;
	for i=#BMTScheduledTasks, 1, -1 do
		entry = BMTScheduledTasks[i];
		if(entry.time <= GetTime()) then
			table.remove(BMTScheduledTasks, i);
			entry.func(unpack(entry));
		end
	end
end

function BMTItem_OnClick(self, button, down)
	-- Function for later use
end

function BMTBottomText_FadingHandler()
	if(BMTFadeState == 0) then
		BMTFadeState = 1;
		BlackMarketTrackerFrameBottomText:SetText(BMT_NEW_VERSION .. " (r" .. BMTHighestRevision .. ")");
		BMTTrackerBottomText_FadeIn();
	elseif(BMTFadeState == 1) then
		BMTFadeState = 2;
		BMTBottomText_FadeOut();
	elseif(BMTFadeState == 2) then
		BMTFadeState = 3;
		local bottomText = "Last updated by ";
		if(BlackMarketTrackerMisc) then
			if(not BlackMarketTrackerMisc.UpdaterName or not BlackMarketTrackerMisc.UpdaterClass or not BlackMarketTrackerMisc.UpdaterGuild or not BlackMarketTrackerMisc.UpdateTime or not BlackMarketTrackerMisc.UpdateDate) then
				bottomText = bottomText .. "|cffffd100Unknown|r";
			else
				local classColor = RAID_CLASS_COLORS[BlackMarketTrackerMisc.UpdaterClass];
				bottomText = bottomText .. "|cff" .. string.format("%2.2x", classColor.r*255) .. string.format("%2.2x", classColor.g*255) .. string.format("%2.2x", classColor.b*255);
				bottomText = bottomText .. BlackMarketTrackerMisc.UpdaterName .. "|r (";
				bottomText = bottomText .. BlackMarketTrackerMisc.UpdaterGuild .. "), ";
				local updateTimeStr =  string.format("%2.2d:%2.2d", (BlackMarketTrackerMisc.UpdateTime/60), (BlackMarketTrackerMisc.UpdateTime%60));
				bottomText = bottomText .. BlackMarketTrackerMisc.UpdateDate .. " " .. updateTimeStr.." (";
				local currHour, currMinute = GetGameTime();
				local currTime = 60*currHour+currMinute;
				local diff = BMT_CalculateDateAndTimeDiff(BlackMarketTrackerMisc.UpdateDate, BlackMarketTrackerMisc.UpdateTime, date("%d/%m/%y"), currTime);
				if(diff > 10080) then
					bottomText = bottomText .. BMT_MORE_THAN_WEEK;
				elseif(diff < 1) then
					bottomText = bottomText .. BMT_LESS_THAN_MINUTE;
				else
					local addSpace = false;
					local days = math.floor(diff/1440);
					local hours = math.floor((diff-(days*1440))/60);
					local minutes = math.floor(diff%60);
					if(days ~= 0) then
						addSpace = true;
						bottomText = bottomText .. tostring(days);
						if(days == 1) then
							bottomText = bottomText .. " day";
						else
							bottomText = bottomText .. " days";
						end						
					end
					
					if(hours ~= 0) then
						if(addSpace) then
							bottomText = bottomText .." ";
						end
						addSpace = true;
						bottomText = bottomText .. tostring(hours);
						if(hours == 1) then
							bottomText = bottomText .. " hour";
						else
							bottomText = bottomText .. " hours";
						end						
					end
					
					if(minutes ~= 0) then
						if(addSpace) then
							bottomText = bottomText .." ";
						end
						addSpace = true;
						bottomText = bottomText .. tostring(minutes);
						if(minutes == 1) then
							bottomText = bottomText .. " minute";
						else
							bottomText = bottomText .. " minutes";
						end						
					end
				end
				bottomText = bottomText .. " ago)";
			end
		else
			bottomText = bottomText .. "|cffffd100Unknown|r";
		end
		BlackMarketTrackerFrameBottomText:SetText(bottomText);
		BMTTrackerBottomText_FadeIn();
	end
end

function BMTTrackerBottomText_FadeIn()
	local fadeInfo = {};
    fadeInfo.mode = "IN";
    fadeInfo.timeToFade = BMT_TEXT_TIME;
	fadeInfo.endAlpha = BMT_TEXT_ALPHA;
    fadeInfo.finishedFunc = BMTBottomText_FadingHandler;
    UIFrameFade(BlackMarketTrackerFrameBottom, fadeInfo);
end

function BMTBottomText_FadeOut()
	local fadeInfo = {};
    fadeInfo.mode = "OUT";
    fadeInfo.timeToFade = BMT_TEXT_TIME;
	fadeInfo.startAlpha = BMT_TEXT_ALPHA;
    fadeInfo.finishedFunc = BMTBottomText_FadingHandler;
    UIFrameFade(BlackMarketTrackerFrameBottom, fadeInfo);
end

function BlackMarketTrackerScrollFrame_Update()
	local numItems = 0;
	if(BlackMarketTrackerCurrentDB ~= nil) then 
		numItems = #BlackMarketTrackerCurrentDB;
	end
	BMT_PredictedTimeRecalculateAll();
	local scrollFrame = BlackMarketTrackerScrollFrame;
	local offset = HybridScrollFrame_GetOffset(scrollFrame);
	local buttons = scrollFrame.buttons;
	local numButtons = #buttons;
	
	for i=1, numButtons do
		local button = buttons[i];
		-- TODO: (should be history db as well) get index of an item in black market
		local index = offset + i;
		if ( index <= numItems ) then
			local name, texture, quantity, itemType, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID = BMT_GetFromCurrentDB(index);
			if(name) then
				button.Name:SetText(name);
                button.Item.IconTexture:SetTexture(texture);
				--[[
                if ( not usable ) then
                    button.Item.IconTexture:SetVertexColor(1.0, 0.1, 0.1);
                else
                    button.Item.IconTexture:SetVertexColor(1.0, 1.0, 1.0);
                end
				]]
                button.Item.Count:SetText(quantity);
                button.Item.Count:SetShown(quantity > 1);
                button.Type:SetText(itemType);
                button.Seller:SetText(sellerName);
                button.Level:SetText(level);
                local bidAmount = currBid;
                local minNextBid = currBid + minIncrement;
                if ( currBid == 0 ) then
                    bidAmount = minBid;
                    minNextBid = minBid;
                end
                MoneyFrame_Update(button.CurrentBid, bidAmount);
                button.minNextBid = minNextBid;
                button.auctionCompleate = (timeLeft == 0);
                button.TimeLeft.Text:SetText(_G["AUCTION_TIME_LEFT"..timeLeft]);
                button.TimeLeft.tooltip = _G["AUCTION_TIME_LEFT"..timeLeft.."_DETAIL"];
                button.itemLink = link;
				local predictedText = BMT_PredictedTimeGetText(index);
				button.PredictedTimeLeft:SetText(predictedText);
				button.Warning:SetText(BMT_COMING_SOON);
                button:Show();
            else
                button:Hide();
			end
		else
			button:Hide();
		end
	end
	
	local totalHeight = numItems * scrollFrame.buttonHeight;
    local displayedHeight = numButtons * scrollFrame.buttonHeight;
    HybridScrollFrame_Update(scrollFrame, totalHeight, displayedHeight);
end

-- should be called only when no new search/sync has been done, as this is throttled and will not update if fields changed but last update was done < 1min ago
-- this will also schedule another refresh
function BMT_FullWindowRefresh()
	BMTTaskManager_Schedule(BMT_WINDOW_REFRESH_DELAY, BMT_FullWindowRefresh);
	if(BlackMarketTrackerCurrentDB == nil) then return end
	if(#BlackMarketTrackerCurrentDB == 0) then return end
	local currHour, currMinute = GetGameTime();
	local currTime = 60*currHour+currMinute;
	if(currTime == BlackMarketTrackerCurrentDB[1].updateTime and date("%d/%m/%y") == BlackMarketTrackerCurrentDB[1].updateDate) then return end
	if(BlackMarketTrackerFrame:IsShown()) then
		BMTFadeState = 2;
		BMTBottomText_FadingHandler();
		BlackMarketTrackerScrollFrame_Update();
	end
end

function BMT_PredictedTimeGetText(index)
	local text = "";
	local predictedLow = BlackMarketTrackerCurrentDB[index].predictedLow;
	local predictedLowH = math.floor(predictedLow/60);
	local predictedLowM = predictedLow%60;
	local predictedHigh = BlackMarketTrackerCurrentDB[index].predictedHigh;
	local predictedHighH = math.floor(predictedHigh/60);
	local predictedHighM = predictedHigh%60;
	-- more than x
	if(predictedHigh > BMT_TIME_LEFT[#BMT_TIME_LEFT-1]) then
		text = "> ";
		if(predictedLowH ~= 0) then
			text = text..tostring(predictedLowH).."h";
		end
		if(predictedLowM ~= 0) then
			if(predictedLowH ~= 0) then text = text .. " "; end
			text = text..tostring(predictedLowM).."m";
		end
	-- less than x
	elseif(predictedLow <= 0 and predictedHigh > 0) then
		text = "< ";
		if(predictedHighH ~= 0) then
			text = text..tostring(predictedHighH).."h";
		end
		if(predictedHighM ~= 0) then
			if(predictedHighH ~= 0) then text = text .. " "; end
			text = text..tostring(predictedHighM).."m";
		end
	-- completed
	elseif(predictedLow <= 0 and predictedHigh <= 0) then
		text = BMT_COMPLETED;
	-- between x and y
	else
		if(predictedLowH ~= 0) then
			text = text..tostring(predictedLowH).."h";
		end
		if(predictedLowM ~= 0) then
			if(predictedLowH ~= 0) then text = text .. " "; end
			text = text..tostring(predictedLowM).."m";
		end
		text = text .. " - ";
		if(predictedHighH ~= 0) then
			text = text..tostring(predictedHighH).."h";
		end
		if(predictedHighM ~= 0) then
			if(predictedHighH ~= 0) then text = text .. " "; end
			text = text..tostring(predictedHighM).."m";
		end
	end
	return text;
end

function BMT_PredictedTimeRecalculateAll()
	if(BlackMarketTrackerCurrentDB == nil) then return end
	if(BMT_PredictedTimeRecalculate == nil) then return end
	for i=1, #BlackMarketTrackerCurrentDB do
		BMT_PredictedTimeRecalculate(i);
	end
end

function BMT_PredictedTimeSetDefaults(index, timeLeft)
	-- completed
	if(timeLeft == 0) then
		BlackMarketTrackerCurrentDB[index]["predictedLow"] = 0;
		BlackMarketTrackerCurrentDB[index]["predictedHigh"] = 0;
	else
		BlackMarketTrackerCurrentDB[index]["predictedLow"] = BMT_TIME_LEFT[timeLeft];
		BlackMarketTrackerCurrentDB[index]["predictedHigh"] = BMT_TIME_LEFT[timeLeft+1];
	end
end

-- recalculates predicted time based just on current time (no new search/sync)
function BMT_PredictedTimeRecalculate(index)
	local currDate = date("%d/%m/%y");
	local currHour, currMinute = GetGameTime();
	local currTime = 60*currHour+currMinute;
	local itemDate = BlackMarketTrackerCurrentDB[index].updateDate;
	local itemTime = BlackMarketTrackerCurrentDB[index].updateTime;
	local diff = BMT_CalculateDateAndTimeDiff(itemDate, itemTime, currDate, currTime);
	-- do nothing if last recalculation was done less than 1 minute ago
	if(diff < 1) then return end;
	local predictedL = BlackMarketTrackerCurrentDB[index].predictedLow;
	local predictedH = BlackMarketTrackerCurrentDB[index].predictedHigh;
	-- do nothing if auction is completed
	if(predictedL <= 0 and predictedH <= 0) then return end
	BlackMarketTrackerCurrentDB[index].updateDate = currDate;
	BlackMarketTrackerCurrentDB[index].updateTime = currTime;
	if(predictedH > BMT_TIME_LEFT[#BMT_TIME_LEFT-1]) then return end
	predictedL = predictedL - diff;
	predictedH = predictedH - diff;
	BlackMarketTrackerCurrentDB[index].predictedLow = predictedL;
	BlackMarketTrackerCurrentDB[index].predictedHigh = predictedH;
end

function BMT_SetNewUpdater(updaterName, updaterClass, updaterGuild, updateTime, updateDate)
	if (not BlackMarketTrackerMisc) then
		BlackMarketTrackerMisc = {};
	else
		-- throttle
		local currHour, currMinute = GetGameTime();
		local currTime = 60*currHour+currMinute;
		if(currTime == BlackMarketTrackerMisc.UpdateTime) then return end
	end
	
	BlackMarketTrackerMisc.UpdaterName = updaterName;
	BlackMarketTrackerMisc.UpdaterClass = updaterClass;
	BlackMarketTrackerMisc.UpdaterGuild = updaterGuild;
	BlackMarketTrackerMisc.UpdateTime = updateTime;
	BlackMarketTrackerMisc.UpdateDate = updateDate;
	
	-- if tracker is opened already, update the bottom text
	if(BlackMarketTrackerFrame:IsShown()) then
		BMTFadeState = 2;
		BMTBottomText_FadingHandler();
	end
end

function BMT_AddNewToCurrentDB(name, texture, quantity, itemType,  level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID)
	-- check if table doesn't exist should already be done
	local currDBSize = #BlackMarketTrackerCurrentDB;
	local currHour, currMinute = GetGameTime();
	local currTime = 60*currHour+currMinute;
	BlackMarketTrackerCurrentDB[currDBSize+1] = {};
	BlackMarketTrackerCurrentDB[currDBSize+1].updateTime = currTime;
	BlackMarketTrackerCurrentDB[currDBSize+1].updateDate = date("%d/%m/%y");
	BlackMarketTrackerCurrentDB[currDBSize+1].name = name;
	BlackMarketTrackerCurrentDB[currDBSize+1].texture = texture;
	BlackMarketTrackerCurrentDB[currDBSize+1].quantity = quantity;
	BlackMarketTrackerCurrentDB[currDBSize+1].itemType = itemType;
	BlackMarketTrackerCurrentDB[currDBSize+1].level = level;
	BlackMarketTrackerCurrentDB[currDBSize+1].levelType = levelType;
	BlackMarketTrackerCurrentDB[currDBSize+1].sellerName = sellerName;
	BlackMarketTrackerCurrentDB[currDBSize+1].minBid = minBid;
	BlackMarketTrackerCurrentDB[currDBSize+1].minIncrement = minIncrement;
	BlackMarketTrackerCurrentDB[currDBSize+1].currBid = currBid;
	-- TODO: Use for highest bidder detection
	BlackMarketTrackerCurrentDB[currDBSize+1].youHaveHighBid = youHaveHighBid;
	-- TODO: Check field
	BlackMarketTrackerCurrentDB[currDBSize+1].numBids = numBids;
	BlackMarketTrackerCurrentDB[currDBSize+1].timeLeft = timeLeft;
	BlackMarketTrackerCurrentDB[currDBSize+1].link = link;
	BlackMarketTrackerCurrentDB[currDBSize+1].marketID = marketID;
	BMT_PredictedTimeSetDefaults(currDBSize+1, timeLeft);
end

function BMT_UpdateExistingCurrentDBEntry(index, name, texture, quantity, itemType, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID)
	local currHour, currMinute = GetGameTime();
	local currTime = 60*currHour+currMinute;
	local prevTimeLeft = BlackMarketTrackerCurrentDB[index].timeLeft;
	-- throttle the updates
	if(prevTimeLeft == timeLeft and currTime == BlackMarketTrackerMisc.UpdateTime and date("%d/%m/%y") == BlackMarketTrackerMisc.UpdateDate) then 
		BMT_Debug("Throttling update. currTime="..currTime.." BMTDB.updateTime="..BlackMarketTrackerMisc.UpdateTime);
		return;
	end
	-- this will just lower the predicted times
	BMT_PredictedTimeRecalculateAll();
	local prevNumBids = BlackMarketTrackerCurrentDB[index].numBids;
	local prevTimeLeft = BlackMarketTrackerCurrentDB[index].timeLeft;
	local prevPredictedLow = BlackMarketTrackerCurrentDB[index].predictedLow;
	local prevPredictedHigh = BlackMarketTrackerCurrentDB[index].predictedHigh;
	local prevPredictedDiff = math.abs(prevPredictedHigh) - math.abs(prevPredictedLow);
	local updateTimeDiff = BMT_CalculateDateAndTimeDiff(BlackMarketTrackerMisc.UpdateDate, BlackMarketTrackerMisc.UpdateTime, date("%d/%m/%y"), currTime);
	-- add time based on number of bids done between the searches
	prevPredictedLow = prevPredictedLow + ((numBids - prevNumBids)*BMT_TIME_EXTEND_PER_BID);
	prevPredictedHigh = prevPredictedHigh + ((numBids - prevNumBids)*BMT_TIME_EXTEND_PER_BID);
	if(timeLeft == 0) then
		BMT_PredictedTimeSetDefaults(index, timeLeft);
	else
		-- time left change up->down
		if(timeLeft < prevTimeLeft) then
			BMT_Debug("index="..index.." UP->DOWN");
			-- time between last update and current time is narrower than time left range given by the game
			if(updateTimeDiff < (BMT_TIME_LEFT[timeLeft+1]-BMT_TIME_LEFT[timeLeft])) then
				BMT_Debug("This is narrower. diff="..updateTimeDiff.." blizz diff="..(BMT_TIME_LEFT[timeLeft+1]-BMT_TIME_LEFT[timeLeft]));
				prevPredictedHigh = BMT_TIME_LEFT[timeLeft+1];
				prevPredictedLow = BMT_TIME_LEFT[timeLeft+1] - updateTimeDiff;
			else
				BMT_Debug("Blizzard's is narrower. diff="..updateTimeDiff.." blizz diff="..(BMT_TIME_LEFT[timeLeft+1]-BMT_TIME_LEFT[timeLeft]));
				BMT_PredictedTimeSetDefaults(index, timeLeft);
			end
		-- change down->up (due to bids)
		elseif(timeLeft > prevTimeLeft) then
			BMT_Debug("index="..index.." DOWN->UP");
			if(updateTimeDiff < (BMT_TIME_LEFT[timeLeft+1]-BMT_TIME_LEFT[timeLeft])) then
				BMT_Debug("This is narrower. diff="..updateTimeDiff.." blizz diff="..(BMT_TIME_LEFT[timeLeft+1]-BMT_TIME_LEFT[timeLeft]));
				prevPredictedLow = BMT_TIME_LEFT[timeLeft];
				prevPredictedHigh= BMT_TIME_LEFT[timeLeft] + updateTimeDiff;
			else
				BMT_Debug("Blizzard's is narrower. diff="..updateTimeDiff.." blizz diff="..(BMT_TIME_LEFT[timeLeft+1]-BMT_TIME_LEFT[timeLeft]));
				BMT_PredictedTimeSetDefaults(index, timeLeft);
			end
		-- narrow the range by cutting come time from the lower predicted range, upper range is already recalculated
		else
			BMT_Debug("index="..index.." UPDATING, NO CHANGE");
			if(prevPredictedLow < BMT_TIME_LEFT[timeLeft]) then
				prevPredictedLow = BMT_TIME_LEFT[timeLeft];
			end
			-- in case this wasn't changed
			if(prevPredictedHigh > BMT_TIME_LEFT[timeLeft+1]) then
				prevPredictedHigh = BMT_TIME_LEFT[timeLeft+1];
			end
		end
		BlackMarketTrackerCurrentDB[index].predictedLow = prevPredictedLow;
		BlackMarketTrackerCurrentDB[index].predictedHigh = prevPredictedHigh;
	end
	BlackMarketTrackerCurrentDB[index].updateTime = currTime;
	BlackMarketTrackerCurrentDB[index].updateDate = date("%d/%m/%y");
	BlackMarketTrackerCurrentDB[index].name = name;
	BlackMarketTrackerCurrentDB[index].texture = texture;
	BlackMarketTrackerCurrentDB[index].quantity = quantity;
	BlackMarketTrackerCurrentDB[index].itemType = itemType;
	BlackMarketTrackerCurrentDB[index].level = level;
	BlackMarketTrackerCurrentDB[index].levelType = levelType;
	BlackMarketTrackerCurrentDB[index].sellerName = sellerName;
	BlackMarketTrackerCurrentDB[index].minBid = minBid;
	BlackMarketTrackerCurrentDB[index].minIncrement = minIncrement;
	BlackMarketTrackerCurrentDB[index].currBid = currBid;
	-- TODO: Use for highest bidder detection
	BlackMarketTrackerCurrentDB[index].youHaveHighBid = youHaveHighBid;
	-- TODO: Check field
	BlackMarketTrackerCurrentDB[index].numBids = numBids;
	BlackMarketTrackerCurrentDB[index].timeLeft = timeLeft;
end

function BMT_GetFromCurrentDB(index)
	return BlackMarketTrackerCurrentDB[index].name, BlackMarketTrackerCurrentDB[index].texture, BlackMarketTrackerCurrentDB[index].quantity, BlackMarketTrackerCurrentDB[index].itemType, BlackMarketTrackerCurrentDB[index].level, BlackMarketTrackerCurrentDB[index].levelType, BlackMarketTrackerCurrentDB[index].sellerName, BlackMarketTrackerCurrentDB[index].minBid, BlackMarketTrackerCurrentDB[index].minIncrement, BlackMarketTrackerCurrentDB[index].currBid, BlackMarketTrackerCurrentDB[index].youHaveHighBid, BlackMarketTrackerCurrentDB[index].numBids, BlackMarketTrackerCurrentDB[index].timeLeft, BlackMarketTrackerCurrentDB[index].link, BlackMarketTrackerCurrentDB[index].marketID, BlackMarketTrackerCurrentDB[index].predictedLow, BlackMarketTrackerCurrentDB[index].predictedHigh;
end

function BMT_UpdateOrAddToCurrentDB(name, texture, quantity, itemType, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID, predictedLow, predictedHigh)
	predictedLow = predictedLow or -1;
	predictedHigh = predictedHigh or -1;
	if(not BlackMarketTrackerCurrentDB) then
		BlackMarketTrackerCurrentDB = {};
	end
	local currDBSize = #BlackMarketTrackerCurrentDB;
	local swapItems = false; -- move old to history and passed to current
	-- iterate through all fields, check for duplicates
	local duplicateIndex = 0;
	for i=1, currDBSize do
		if(BlackMarketTrackerCurrentDB[i].marketID == marketID) then
			duplicateIndex = i;
			local currHour, currMinute = GetGameTime();
			local currTime = 60*currHour+currMinute;
			if(BlackMarketTrackerCurrentDB[i].name ~= name or (BMT_CalculateDateAndTimeDiff(BlackMarketTrackerCurrentDB[i].updateDate, BlackMarketTrackerCurrentDB[i].updateTime, date("%d/%m/%y"), currTime) > BMT_ITEM_REMOVE_TIME_THRESHOLD)) then
				swapItems = true;
			end
		end
	end
	-- new entry
	if(duplicateIndex == 0) then
		BMT_AddNewToCurrentDB(name, texture, quantity, itemType, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID);
	-- same marketIDs but different names or old item still sitting on marketID that has been used again already
	elseif(duplicateIndex ~= 0 and swapItems) then
		BMT_MoveEntryToHistoryDB(duplicateIndex);
		BMT_AddNewToCurrentDB(name, texture, quantity, itemType, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID);
	-- existing entry
	else
		BMT_UpdateExistingCurrentDBEntry(duplicateIndex, name, texture, quantity, itemType, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID);
	end
end

function BMT_MoveEntryToHistoryDB(index)
	if(not BlackMarketTrackerHistoryDB) then
		BlackMarketTrackerHistoryDB = {};
	end
	BMT_Debug("Moving "..BlackMarketTrackerCurrentDB[index].name.." to history db");
	local histDBSize = #BlackMarketTrackerHistoryDB;
	BlackMarketTrackerHistoryDB[histDBSize+1] = {};
	BlackMarketTrackerHistoryDB[histDBSize+1].name = BlackMarketTrackerCurrentDB[index].name;
	BlackMarketTrackerHistoryDB[histDBSize+1].texture = BlackMarketTrackerCurrentDB[index].texture;
	BlackMarketTrackerHistoryDB[histDBSize+1].quantity = BlackMarketTrackerCurrentDB[index].quantity;
	BlackMarketTrackerHistoryDB[histDBSize+1].itemType = BlackMarketTrackerCurrentDB[index].itemType;
	BlackMarketTrackerHistoryDB[histDBSize+1].level = BlackMarketTrackerCurrentDB[index].level;
	BlackMarketTrackerHistoryDB[histDBSize+1].levelType = BlackMarketTrackerCurrentDB[index].levelType;
	BlackMarketTrackerHistoryDB[histDBSize+1].sellerName = BlackMarketTrackerCurrentDB[index].sellerName;
	BlackMarketTrackerHistoryDB[histDBSize+1].minBid = BlackMarketTrackerCurrentDB[index].minBid;
	BlackMarketTrackerHistoryDB[histDBSize+1].minIncrement = BlackMarketTrackerCurrentDB[index].minIncrement;
	BlackMarketTrackerHistoryDB[histDBSize+1].currBid = BlackMarketTrackerCurrentDB[index].currBid;
	BlackMarketTrackerHistoryDB[histDBSize+1].youHaveHighBid = BlackMarketTrackerCurrentDB[index].youHaveHighBid;
	BlackMarketTrackerHistoryDB[histDBSize+1].numBids = BlackMarketTrackerCurrentDB[index].numBids;
	BlackMarketTrackerHistoryDB[histDBSize+1].link = BlackMarketTrackerCurrentDB[index].link;
	BlackMarketTrackerHistoryDB[histDBSize+1].addedDate = date("%d/%m/%y");
	table.remove(BlackMarketTrackerCurrentDB, index);
end

-- verifies if given number of fields can be sent through the SendAddonMessage() with given db index and starting field number. Returns the end field number.
-- item sync messages in format: BMT_SYNC_ITEM_KEY%marketID%startField%endField%Field(startField)%Field(startField+1)%...%Field(endField)
-- where % is the separator delimiting the fields
function BMT_CalculateSyncFields(index, startField)
	local numFields = #BMTItemSyncFields;
	local length = string.len(BMT_SYNC_ITEM_KEY) + string.len(tostring((BlackMarketTrackerCurrentDB[index].marketID))) + string.len(tostring(startField)) + 3*string.len(BMT_SEPARATOR);
	local endField;
	for endField=startField, numFields do
		-- add end field number string
		length = length + string.len(tostring(endField));
		-- add separator
		length = length + string.len(BMT_SEPARATOR);
		-- add field contents
		length = length + string.len(tostring(BlackMarketTrackerCurrentDB[index][BMTItemSyncFields[endField]]));
		-- string too long
		if(length > BMT_MAX_ADDON_MSG_LEN) then
			if(endField == startField) then
				-- not even one field can be sent
				return -1;
			else
				return (endField-1);
			end
		end
		-- do this before the next iteration
		length = length - string.len(tostring(endField));
	end
	
	return numFields;
end

-- TODO: Friends
function BMT_SendCurrentDB()
	if(not BlackMarketTrackerCurrentDB or not BlackMarketTrackerMisc) then return end
	if(not BlackMarketTrackerMisc.UpdaterName) then return end
	
	local currDBSize = #BlackMarketTrackerCurrentDB;
	local numFields = #BMTItemSyncFields;
	for i=1, currDBSize do
		BMT_PredictedTimeRecalculateAll();
		local name, texture, quantity, itemType, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID, predictedLow, predictedHigh = BMT_GetFromCurrentDB(i);
		local startField = 1;
		while(startField <= numFields) do
			local endField = BMT_CalculateSyncFields(i, startField);
			if(endField == -1) then 
				BMT_Debug("Unable to send field="..startField.." index="..index.." value="..BlackMarketTrackerCurrentDB[index][BMTItemSyncFields[startField]]);
				break;
			end
			local msg = BMT_SYNC_ITEM_KEY..BMT_SEPARATOR..marketID..BMT_SEPARATOR..tostring(startField)..BMT_SEPARATOR..tostring(endField);
			for j=startField, endField do
				msg = msg..BMT_SEPARATOR..BlackMarketTrackerCurrentDB[i][BMTItemSyncFields[j]];
			end
			startField = endField+1;
			SendAddonMessage("BMT", msg, "GUILD");
		end
	end
	SendAddonMessage("BMT", BMT_SYNC_ITEMEND_KEY, "GUILD");
	SendAddonMessage("BMT", BMT_SYNC_SENDER_KEY..BMT_SEPARATOR..BlackMarketTrackerMisc.UpdaterName..BMT_SEPARATOR..BlackMarketTrackerMisc.UpdaterClass..BMT_SEPARATOR..BlackMarketTrackerMisc.UpdaterGuild..BMT_SEPARATOR..BlackMarketTrackerMisc.UpdateTime..BMT_SEPARATOR..BlackMarketTrackerMisc.UpdateDate, "GUILD");
end

function BMT_NewSearchHandler()
	-- check if blizzard's black market ui is still opened
	if(not BlackMarketFrame:IsShown()) then return end
	local playerName = UnitName("player");
	local _, playerClass = UnitClass("player");
	local playerGuild = GetGuildInfo("player");
	local updateHour, updateMinute = GetGameTime();
	local updateTime = 60*updateHour+updateMinute;
	local updateDate = date("%d/%m/%y");
	local numItems = C_BlackMarket.GetNumItems();
	if(numItems == nil) then return end
	for i=1, numItems do
		local name, texture, quantity, itemType, _, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID = C_BlackMarket.GetItemInfoByIndex(i)
		BMT_UpdateOrAddToCurrentDB(name, texture, quantity, itemType, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID);
	end
	-- move to history DB
	if(BlackMarketTrackerCurrentDB) then
		local currDBSize = #BlackMarketTrackerCurrentDB;
		-- iterate through all items in db
		local found;
		local i=1;
		while(i<=currDBSize) do
			found = 0;
			-- iterate through bmah items
			for j=1, numItems do
				local name, texture, quantity, itemType, _, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID = C_BlackMarket.GetItemInfoByIndex(j);
				if(BlackMarketTrackerCurrentDB[i].marketID == marketID) then
					found = 1;
				end
			end
			if(found == 0) then
				BMT_MoveEntryToHistoryDB(i);
				i = i-1;
				currDBSize = currDBSize-1;
			end
			i = i+1;
		end
	end
	BMT_SetNewUpdater(playerName, playerClass, playerGuild, updateTime, updateDate);
	-- update the list in case the addon's browser is opened
	if(BlackMarketTrackerFrame:IsShown()) then
		BMTFadeState = 2;
		BMTBottomText_FadingHandler();
		BlackMarketTrackerScrollFrame_Update();
	end
	-- sync to others
	BMT_SendCurrentDB();
end

function BMT_SyncSenderUpdateHandler(msg)
	local _, updaterName, updaterClass, updaterGuild, updateTime, updateDate = strsplit(BMT_SEPARATOR, msg);
	
	if(BMTSyncSender == nil) then
		BMTSyncSender = {};
	end
	BMTSyncSender.UpdaterName = updaterName;
	BMTSyncSender.UpdaterClass = updaterClass;
	BMTSyncSender.UpdaterGuild = updaterGuild;
	BMTSyncSender.UpdateTime = tonumber(updateTime);
	BMTSyncSender.UpdateDate = updateDate;
	BMTTaskManager_Unschedule(BMT_SYNC_ITEMS_DELAY, BMT_FinishSyncingItems);
	BMTTaskManager_Schedule(BMT_SYNC_ITEMS_DELAY, BMT_FinishSyncingItems);
end

function BMT_SyncItemUpdateHandler(msg)
	-- sync_key%markedID%startField%endField%field1%field2%...
	local fields = {strsplit(BMT_SEPARATOR, msg)};
	-- refresh the delay
	BMTTaskManager_Unschedule(BMT_SYNC_ITEMS_DELAY, BMT_FinishSyncingItems);
	BMTTaskManager_Schedule(BMT_SYNC_ITEMS_DELAY, BMT_FinishSyncingItems);
	if(not BMTSyncItemDB) then
		BMTSyncItemDB = {};
	end
	local marketID = tonumber(fields[2]);
	local startField = tonumber(fields[3]);
	local endField = tonumber(fields[4]);
	BMT_Debug("startField="..startField.." endField="..endField);
	-- first field is at fields[5]
	if(not BMTSyncItemDB[marketID]) then
		BMTSyncItemDB[marketID] = {};
	end
	for i=1, (#fields-4) do
		-- TODO: Check if this has narrower preticted times, if yes - update based on
		-- passed Blizzard's timeLeft and own predicted times. Resync it back later (no updater)
		BMT_Debug("i+4="..(i+4).." fields[i+4]="..fields[i+4].." BMTItemSyncFields[i+startField-1]="..BMTItemSyncFields[i+startField-1]);
		if(BMTIsSyncFieldNumber[i+startField-1]) then
			BMTSyncItemDB[marketID][BMTItemSyncFields[i+startField-1]] = tonumber(fields[i+4]);
		else
			BMTSyncItemDB[marketID][BMTItemSyncFields[i+startField-1]] = fields[i+4];
		end
	end
end

function BMT_FinishSyncingItems()
	if(not BMTSyncItemDB) then return end
	if(not BlackMarketTrackerCurrentDB) then
		BlackMarketTrackerCurrentDB = {};
	end
	local currDBSize = #BlackMarketTrackerCurrentDB;
	-- update already existing or add new to current db
	--for i=1, syncDBSize do
	for marketID in pairs(BMTSyncItemDB) do
		BMT_UpdateOrAddToCurrentDB(BMTSyncItemDB[marketID].name, BMTSyncItemDB[marketID].texture, BMTSyncItemDB[marketID].quantity, BMTSyncItemDB[marketID].itemType, BMTSyncItemDB[marketID].level, BMTSyncItemDB[marketID].levelType, BMTSyncItemDB[marketID].sellerName, BMTSyncItemDB[marketID].minBid, BMTSyncItemDB[marketID].minIncrement, BMTSyncItemDB[marketID].currBid, BMTSyncItemDB[marketID].youHaveHighBid, BMTSyncItemDB[marketID].numBids, BMTSyncItemDB[marketID].timeLeft, BMTSyncItemDB[marketID].link, marketID, BMTSyncItemDB[marketID].predictedLow, BMTSyncItemDB[marketID].predictedHigh);
		local size = #BMTSyncItemDB[marketID];
		BMT_Debug("size="..size);
	end
	-- remove old from history db
	local i=1;
	while(i<=currDBSize) do
		found = 0;
		-- iterate through synced items
		for marketID in pairs(BMTSyncItemDB) do
			if(BlackMarketTrackerCurrentDB[i].marketID == marketID) then
				found = 1;
			end
		end
		if(found == 0) then
			BMT_MoveEntryToHistoryDB(i);
			i = i-1;
			currDBSize = currDBSize-1;
		end
		i = i+1;
	end
	if(BMTSyncSender ~= nil) then
		BMT_SetNewUpdater(BMTSyncSender.UpdaterName, BMTSyncSender.UpdaterClass, BMTSyncSender.UpdaterGuild, BMTSyncSender.UpdateTime, BMTSyncSender.UpdateDate);
	end
	BMTSyncItemDB = nil;
	BMTSyncSender = nil;
	if(BlackMarketTrackerFrame:IsShown()) then
		BlackMarketTrackerScrollFrame_Update();
	end
end

function BMT_SyncVersionHandler(msg)
	local key, verNumberStr = strsplit(BMT_SEPARATOR, msg);
	local verNum = tonumber(verNumberStr);
	-- this is higher version
	if(BMTRevision > verNum and key == BMT_SYNC_VERREQ_KEY) then
		SendAddonMessage("BMT", BMT_SYNC_VERRESP_KEY..BMT_SEPARATOR..BMTRevision, "GUILD");
	elseif(BMTRevision < verNum) then
		BMTHighestRevision = verNum;
		-- if ui is open, update the bottom text
		if(BlackMarketTrackerFrame:IsShown()) then
			BMTFadeState = 0;
			BMTBottomText_FadingHandler();
		end
	end
end

-- returns difference between given poits in time in minutes (absolute)
-- date is a string in "dd/mm/yy" format, time is a number in 60*hour+minutes format
function BMT_CalculateDateAndTimeDiff(firstDate, firstTime, secondDate, secondTime)
	local t = {};
	local day, month, year = strsplit("/", firstDate);
	t["year"] = tonumber(year) + 2000;
	t["month"] = tonumber(month);
	t["day"] = tonumber(day);
	t["hour"] = firstTime / 60;
	t["min"] = firstTime % 60;
	t["sec"] = 0;
	local first = time(t);
	day, month, year = strsplit("/", secondDate);
	t["year"] = tonumber(year) + 2000;
	t["month"] = tonumber(month);
	t["day"] = tonumber(day);
	t["hour"] = secondTime / 60;
	t["min"] = secondTime % 60;
	t["sec"] = 0;
	local second = time(t);
	return (math.floor(math.abs(first-second)/60));
end

-- returns "smaller" if first date/time is smaller, "equal" if they are equal, "greater" if first is greater.
-- date is a string in "dd/mm/yy" format, time is a number in 60*hour+minutes format
function BMT_CompareDateAndTime(firstDate, firstTime, secondDate, secondTime)
	local firstDay, firstMonth, firstYear = strsplit("/", firstDate);
	local secondDay, secondMonth, secondYear = strsplit("/", secondDate);
	if(firstYear > secondYear) then return "greater"
	elseif(firstYear < secondYear) then return "smaller"
	else
		if(firstMonth > secondMonth) then return "greater"
		elseif(firstMonth < secondMonth) then return "smaller"
		else
			if(firstDay > secondDay) then return "greater"
			elseif(firstDay < secondDay) then return "smaller"
			else
				if(firstTime > secondTime) then return "greater"
				elseif(firstTime < secondTime) then return "smaller"
				else return "equal"
				end
			end
		end
	end
end

-- response to update time request
function BMT_SyncSendUpdateTime(msg)
	-- check if this has db
	if(not BlackMarketTrackerMisc or not BlackMarketTrackerCurrentDB) then
		SendAddonMessage("BMT", BMT_SYNC_TIMERESP_KEY..BMT_SEPARATOR..BMTBestUpdateDate..BMT_SEPARATOR..BMTBestUpdateTime, "GUILD");
		return;
	end
	local _, updateDate, updateTime = strsplit(BMT_SEPARATOR, msg);
	if(BMT_CompareDateAndTime(updateDate, tonumber(updateTime), BlackMarketTrackerMisc.UpdateDate, BlackMarketTrackerMisc.UpdateTime) == "smaller") then
		-- this has newer database, let others know
		SendAddonMessage("BMT", BMT_SYNC_TIMERESP_KEY..BMT_SEPARATOR..BlackMarketTrackerMisc.UpdateDate..BMT_SEPARATOR..BlackMarketTrackerMisc.UpdateTime, "GUILD");
	end
end

-- received update time
function BMT_SyncUpdateTimeReceiveHandler(msg, sender)
	local _, updateDate, updateTime = strsplit(BMT_SEPARATOR, msg);
	if(BMT_CompareDateAndTime(updateDate, tonumber(updateTime), BMTBestUpdateDate, BMTBestUpdateTime) == "greater") then
		BMTBestUpdateDate = updateDate;
		BMTBestUpdateTime = updateTime;
		BMTBestUpdateTimeSender = sender;
	end
end

function BMT_AddonMsgReceiveHandler(prefix, msg, channel, sender)
	if(prefix ~= "BMT") then return end
	BMT_Debug("("..sender..") msg="..msg);
	if(sender == UnitName("player")) then return end
	local key = strsplit(BMT_SEPARATOR, msg);
	if(key == BMT_SYNC_SENDER_KEY) then
		BMT_SyncSenderUpdateHandler(msg);
	elseif(key == BMT_SYNC_ITEM_KEY) then
		BMT_SyncItemUpdateHandler(msg);
	elseif(key == BMT_SYNC_VERREQ_KEY or key == BMT_SYNC_VERRESP_KEY) then
		BMT_SyncVersionHandler(msg);
	elseif(key == BMT_SYNC_TIMEREQ_KEY) then
		BMT_SyncSendUpdateTime(msg);
	elseif(key == BMT_SYNC_TIMERESP_KEY and BMTSyncingTime == true) then
		BMT_SyncUpdateTimeReceiveHandler(msg, sender);
	elseif(key == BMT_SYNC_ITEMREQ_KEY) then
		local _, name = strsplit(BMT_SEPARATOR, msg);
		if(name == UnitName("player")) then
			BMT_SendCurrentDB();
		end
	elseif(key == BMT_SYNC_ITEMEND_KEY) then
		-- don't request again, which would happen if more people logged in at the same time
		BMTTaskManager_Unschedule(BMT_FinishSyncingTime);
		BMTTaskManager_Unschedule(BMT_FinishSyncingItems);
		BMT_FinishSyncingItems();
	end
end

function BMT_FinishSyncingTime()
	-- ignore other responses
	BMTSyncingTime = false;
	-- check if someone had better DB
	if(BMTBestUpdateTimeSender == UnitName("player") or BMTBestUpdateTimeSender == nil) then return end
	-- request best database
	SendAddonMessage("BMT", BMT_SYNC_ITEMREQ_KEY..BMT_SEPARATOR..BMTBestUpdateTimeSender, "GUILD");
end

function BMT_OnLoginHandler()
	-- sync version
	SendAddonMessage("BMT", BMT_SYNC_VERREQ_KEY..BMT_SEPARATOR..BMTRevision, "GUILD");
	BMTSyncingTime = true;
	-- get best times
	if(BlackMarketTrackerMisc) then
		if(BlackMarketTrackerMisc.UpdateTime) then
			BMTBestUpdateDate = BlackMarketTrackerMisc.UpdateDate;
			BMTBestUpdateTime = BlackMarketTrackerMisc.UpdateTime;
			BMTBestUpdateTimeSender = UnitName("player");
		end
	end
	SendAddonMessage("BMT", BMT_SYNC_TIMEREQ_KEY..BMT_SEPARATOR..BMTBestUpdateDate..BMT_SEPARATOR..BMTBestUpdateTime, "GUILD");
	BMTTaskManager_Schedule(BMT_SYNC_TIME_DELAY, BMT_FinishSyncingTime);
end